/* YAK - Copyright (c) 1997 Timo Sirainen - read license.txt */

/* files.c - File handling */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>

#include "os.h"
#include "memory.h"
#include "files.h"
#include "output.h"
#include "logfile.h"
#include "timeslic.h"

static unsigned long fgets_start, fgets_end;
static char *fgets_buf;
static int fgets_buf_len, fgets_handle;

#ifdef DEBUG_FILES
typedef struct FHANDLE_REC
{
    struct FHANDLE_REC *next;

    int handle;
    unsigned long linenum;
    char *openname;
#ifdef __WATCOMC__
    char fname[];
#else
    char fname[0];
#endif
}
FHANDLE_REC;

static FHANDLE_REC *first, *last;
#endif

void _files_init(void)
{
#ifdef DEBUG_FILES
    first = NULL; last = NULL;
#endif
    fgets_handle = -1; fgets_buf = NULL; fgets_buf_len = 0;
}

void _files_deinit(void)
{
#ifdef DEBUG_FILES
    FHANDLE_REC *rec;

    while (first != NULL)
    {
        write_log("Unclosed file handle for file %s (in file %s line %lu)",
                  first->openname == NULL ? "(null)" : first->openname,
                  first->fname, first->linenum);
        _free(first->openname);
        rec = first->next;
        _free(first);
        first = rec;
    }
#endif

    if (fgets_buf != NULL) _free(fgets_buf);
}

#ifdef DEBUG_FILES
int _FileCreate(char *openname, int flags, char *fname, unsigned long linenum)
#else
int _FileCreate(char *openname, int flags)
#endif
{
#ifdef DEBUG_FILES
    FHANDLE_REC *rec;
#endif
    int ret;

    ret = creat(openname, flags);

    if (ret == -1) return -1;

#ifdef DEBUG_FILES
    /* Add to linked list */
    rec = (FHANDLE_REC *) _malloc(sizeof(FHANDLE_REC)+strlen(fname)+1);
    if (rec == NULL)
    {
        write_log("FileOpen() : Not enough memory!");
        close(ret);
        return -1;
    }

    if (first == NULL) first = rec;
    if (last != NULL) last->next = rec;
    rec->next = NULL; last = rec;

    rec->openname = _malloc(strlen(openname)+1);
    if (rec->openname != NULL) strcpy(rec->openname, openname);
    rec->handle = ret;
    rec->linenum = linenum;
    strcpy(rec->fname, fname);
#endif

    return ret;
}

#ifdef DEBUG_FILES
#  ifdef UNIX
   int _FileOpen(char *openname, int flags, char *fname, unsigned long linenum)
#  else
   int _FileOpen(char *openname, int flags, int share, char *fname, unsigned long linenum)
#  endif
#else
#  ifdef UNIX
   int _FileOpen(char *openname, int flags)
#  else
   int _FileOpen(char *openname, int flags, int share)
#  endif
#endif
{
#ifdef DEBUG_FILES
    FHANDLE_REC *rec;
#endif
    int ret;

    /* Open file */
#ifdef UNIX
    ret = open(openname, flags);
#else
    ret = sopen(openname, flags, share);
#endif

    if (ret == -1) return -1;

#ifdef DEBUG_FILES
    /* Add to linked list */
    rec = (FHANDLE_REC *) _malloc(sizeof(FHANDLE_REC)+strlen(fname)+1);
    if (rec == NULL)
    {
        write_log("FileOpen() : Not enough memory!");
        close(ret);
        return -1;
    }

    if (first == NULL) first = rec;
    if (last != NULL) last->next = rec;
    rec->next = NULL; last = rec;

    rec->openname = _malloc(strlen(openname)+1);
    if (rec->openname != NULL) strcpy(rec->openname, openname);
    rec->handle = ret;
    rec->linenum = linenum;
    strcpy(rec->fname, fname);
#endif

    return ret;
}

#ifdef DEBUG_FILES
void _FileClose(int f, char *fname, unsigned long linenum)
#else
void _FileClose(int f)
#endif
{
#ifdef DEBUG_FILES
    FHANDLE_REC *rec, *prev;
#endif

    if (fgets_handle == f) fgets_handle = -1;

#ifdef DEBUG_FILES
    rec = first; prev = NULL;
    while (rec != NULL)
    {
        if (rec->handle == f)
        {
            if (last == rec)
            {
                last = prev;
                if (last != NULL) last->next = NULL;
            }
            if (prev != NULL) prev->next = rec->next; else first = rec->next;
            if (rec->openname != NULL) _free(rec->openname);
            _free(rec);
            close(f);
            return;
        }
        prev = rec; rec = rec->next;
    }

    if (rec == NULL)
        write_log("Tried to close unopened file handle %d, file %s line %lu", f, fname, linenum);
#endif
    close(f);
}

char *_fgets(char *buf, int len, int handle)
{
    char *strp;
    int readed, left;
    unsigned long fpos;

    fpos = FileSeek(handle, 0, SEEK_CUR);

    left = 0;
    if (fgets_handle == handle)
    {
        /* Get as much data from buffer as possible */
        if (fgets_start <= fpos && fgets_end >= fpos)
        {
            left = (fgets_end-fpos) > len ? len : fgets_end-fpos;
            memcpy(buf, fgets_buf + (fpos-fgets_start), left);
            len -= left;
            FileSeek(handle, left, SEEK_CUR);

            for (strp = buf+left; strp > buf; strp--)
                if (*strp == '\r' || *strp == '\n' || *strp == 26)
                {
                    readed = left;
                    goto __argh_i_used_goto;
                }
        }
    }

    readed = FileRead(handle, buf+left, len)+left;
    if (readed-left <= 0) return NULL;

    /* Save to buffer */
    if (fgets_buf_len < readed-left)
    {
        _free(fgets_buf);
        fgets_buf = (char *) _malloc(readed-left);
        if (fgets_buf == NULL) fgets_buf_len = 0;
    }
    if (fgets_buf != NULL)
    {
        fgets_start = fpos+left;
        fgets_end = fpos+readed-left;
        memcpy(fgets_buf, buf+left, readed-left);
    }

__argh_i_used_goto:
    if (*buf == 26) return NULL; /* EOF */

    strp = buf; left = readed;
    while (left != 0)
    {
        left--;
        if (*strp == '\r' || *strp == '\n' || *strp == 26)
        {
            if (*strp == '\r' && left > 0 && *(strp+1) == '\n') left--;
            *strp = '\0';
            break;
        }
        strp++;
    }

    if (left != 0) FileSeek(handle,fpos+readed-left,SEEK_SET);

    return buf;
}

int copyfile(char *src, char *dest)
{
    FILE *Fsrc,*Fdst;
    char *buf;
    size_t readed;

    if ((Fsrc = fopen(src,"rb")) == NULL) return 0;
    if ((Fdst = fopen(dest,"w+b")) == NULL) { fclose(Fsrc); return 0; }
    chmod(dest, 0660);
    buf = (char *) _malloc(4096);

    do {
        readed = fread(buf,1,4096,Fsrc);
        fwrite(buf,1,readed,Fdst);
    } while (readed == 4096);

    _free(buf);
    fclose(Fsrc); fclose(Fdst);
    return 1;
}

int FileLock(int handle)
{
    int tries;

    tries = 0;
    while (lock(handle,0,1) != 0)
    {
#ifdef UNIX
        if (errno != EWOULDBLOCK)
        {
#ifdef __OUTPUT_H
            output("\rFATAL ERROR while locking file!!!\r");
            write_log("FATAL ERROR while locking file!!!");
#endif
            return 0;
        }
#endif

        tries++;
#ifdef __OUTPUT_H
        if (tries == 20)
        {
            output("\rPlease wait, trying to lock file (takes max. 20 sec)...\r");
        }
        if (tries == 60)
        {
            output("\rOK, I give up. Something has to be wrong..\r\n");
            write_log("Can't lock file!\n");
            return 0;
        }
        usleep(500000);
#endif
        give_timeslice();
    }

    return 1;
}

#ifndef __NO_DRIVES__
int change_dir(char *path)
{
#ifndef __EMX__
    unsigned drives;
#endif

    if (path[0] == '\0') return 0;
    if (path[1] == ':')
    {
#ifdef __EMX__
        _chdrive(path[0]);
#else
        _dos_setdrive(toupper(path[0])-'A'+1,&drives);
#endif
        path += 2;
    }
    return chdir(path);
}
#endif

#ifdef __FILES_CASE_SENSITIVE__
char *search_ign_file(char *fname)
{
    DIR *dirp;
    struct dirent *direntp;
    char *strp;

    /* Open directory */
    strp = strrchr(fname,SLASH);
    if (strp != NULL)
    {
        *strp = '\0';
        dirp = opendir(fname);
        *strp++ = SLASH;
    }
    else
    {
        strp = fname;
        dirp = opendir(".");
    }

    if (dirp == NULL) return NULL;

    /* Read directory */
    for (;;)
    {
        direntp = readdir(dirp);
        if (direntp == NULL) break;

        if (stricmp(direntp->d_name, strp) == 0)
        {
            /* Found file! */
            strcpy(strp, direntp->d_name);
            closedir(dirp);
            return fname;
        }
    }
    closedir(dirp);

    return NULL;
}

int FileOpenIgn(char *fname, int flags, int share)
{
    char str[MAX_PATH_LEN];

    share = share;
    strcpy(str,fname);
    if (search_ign_file(str) == NULL) return -1;

    return FileOpen(str,flags,0);
}
#endif

int mkpath(char *path)
{
    char str[256];
#ifndef __NO_DRIVES__
    unsigned drive;
#endif
    int pos,slen,spos;
    int start;
    struct stat statbuf;

    slen = strlen(path);

    if (path[slen-1] == SLASH) slen--;

    start = 0;
#ifndef __NO_DRIVES__
    if (path[1] == ':')
    {
        drive = toupper(path[0]);
        start = 2;
    }
    else
    {
#ifdef __EMX__
        drive = _getdrive();
#else
        _dos_getdrive(&drive); drive += 'A'-1;
#endif
    }

    if (path[start] == SLASH)
    {
        str[0] = (char) drive;
        str[1] = ':';
        str[2] = SLASH;
        str[3] = '\0';
        start++;
    } else str[0] = '\0';
#else
    if (path[0] == SLASH)
    {
        str[0] = SLASH;
        start++;
    }
    str[start] = '\0';
#endif

    spos = strlen(str);
    do
    {
        for (pos=start; pos<slen; pos++)
        {
            if (path[pos] == SLASH) break;

            str[spos] = path[pos];
            spos++;
        }
        str[spos] = 0;

        start = pos+1;
        if (stat(str, &statbuf) != 0)
        {
#ifdef UNIX
            if (mkdir(str, 0770) != 0) return 0;
            chmod(str, 0770); /* Why doesn't Linux give group-write access just with mkdir() ?? */
#elif defined (__EMX__)
            if (mkdir(str, 0) != 0) return 0;
#else
            if (mkdir(str) != 0) return 0;
#endif
        }

        str[spos] = SLASH; str[spos+1] = 0; spos++;
    } while (slen >= start);

    return 1;
}

int insert_space(int handle, long pos, long size, long fsize)
{
    char tmpbuf[128];
    char *buffer;

    unsigned bufsize;
    unsigned long movesz,lpos;

    bufsize = 16384;
    movesz = fsize - pos;
    if (bufsize > movesz) bufsize = (unsigned) movesz;
    if (bufsize == 0) return 0;

    if ((bufsize <= sizeof(tmpbuf)) || ((buffer = (char *) _malloc(bufsize)) == NULL))
    {
        bufsize = sizeof(tmpbuf);
        buffer = tmpbuf;
    }

    for (lpos=1; lpos<=movesz/bufsize; lpos++)
    {
        FileSeek(handle,fsize-bufsize*lpos,SEEK_SET);
        if (FileRead(handle,buffer,bufsize) != (int) bufsize)
        {
            if (buffer != tmpbuf) _free(buffer);
            return 0;
        }
        FileSeek(handle,fsize-bufsize*lpos+size,SEEK_SET);
        if (FileWrite(handle,buffer,bufsize) != (int) bufsize)
        {
            if (buffer != tmpbuf) _free(buffer);
            return 0;
        }
    }

    if (movesz % bufsize > 0)
    {
        FileSeek(handle,pos,SEEK_SET);
        if (FileRead(handle,buffer,(unsigned) (movesz % bufsize)) != (int) (movesz % bufsize))
        {
            if (buffer != tmpbuf) _free(buffer);
            return 0;
        }
        FileSeek(handle,pos+size,SEEK_SET);
        if (FileWrite(handle,buffer,(unsigned) (movesz % bufsize)) != (int) (movesz % bufsize))
        {
            if (buffer != tmpbuf) _free(buffer);
            return 0;
        }
    }

    if (buffer != tmpbuf) _free(buffer);
    return 1;
}

FILE_BUF *FileBufOpen(char *fname, int mode, int share_mode)
{
    FILE_BUF *file;
    int handle;
    struct stat statbuf;

    if (stat(fname, &statbuf) != 0) return NULL;

    handle = FileOpen(fname, mode, share_mode);
    if (handle == -1) return NULL;

    file = (FILE_BUF *) _malloc(sizeof(FILE_BUF));
    if (file == NULL)
    {
        FileClose(handle);
        return NULL;
    }

    file->handle = handle;
    file->last_mtime = statbuf.st_mtime;

    file->buffering = 1;
    file->bufstart = 1;
    file->bufend = 0;

    file->seekpos = 0;
    file->filesize = lseek(handle, 0, SEEK_END);
    lseek(handle, 0, SEEK_SET);

    return file;
}

FILE_BUF *FileBufCreate(char *fname, int mode)
{
    FILE_BUF *file;
    int handle;

    handle = FileCreate(fname, mode);
    if (handle == -1) return NULL;

    file = (FILE_BUF *) _malloc(sizeof(FILE_BUF));
    if (file == NULL)
    {
        FileClose(handle);
        return NULL;
    }

    file->handle = handle;

    file->buffering = 1;
    file->bufstart = 1;
    file->bufend = 0;

    file->seekpos = 0;
    file->filesize = lseek(handle, 0, SEEK_END);
    lseek(handle, 0, SEEK_SET);

    return file;
}

void FileBufClose(FILE_BUF *file)
{
    if (file == NULL) return;

    FileClose(file->handle);
    _free(file);
}

void FileBufClear(FILE_BUF *file)
{
    file->bufstart = 1;
    file->bufend = 0;
    file->filesize = lseek(file->handle, 0, SEEK_END);
}

int FileBufRead(FILE_BUF *file, void **buffer, int bufsize)
{
    int readed,pos;
    long fpos;

#if 0
    struct stat statbuf;

    if (fstat(file->handle, &statbuf) != 0) return 0;

    if (statbuf.st_mtime != file->last_mtime)
    {
        /* File modified, dump buffer */
        file->last_mtime = statbuf.st_mtime;
        file->bufstart = 1;
        file->bufend = 0;
        file->filesize = lseek(file->handle, 0, SEEK_END);
    }
#endif

    if (bufsize > (int) sizeof(file->buffer))
    {
        printf("ReadBufFile() : Buffer too small! This should NEVER happen!!\n");
        write_log("ReadBufFile() : Buffer too small! This should NEVER happen!!");
        return 0;
    }

    if (file->seekpos >= (long) file->filesize) return 0;

    if (file->bufstart > file->seekpos || (file->bufend < file->seekpos+bufsize-1 && (unsigned long) file->bufend != file->filesize-1))
    {
        /*if (file->bufstart > file->seekpos && file->buffering)
        {*/
            /* Probably reading backwards.. */
        /*    fpos = file->seekpos-sizeof(file->buffer)+bufsize+256;
            if (fpos > file->seekpos) fpos = file->seekpos;
            if (fpos < 0) fpos = 0;
        }
        else
        {*/
            /* Probably seeking forward */
            fpos = file->seekpos;
        //}
        FileSeek(file->handle, fpos, SEEK_SET);
        readed = FileRead(file->handle, file->buffer, file->buffering ? sizeof(file->buffer) : bufsize);
        if (readed < 1) return -1;

        file->bufstart = fpos;
        file->bufend = file->bufstart+readed-1;

        pos = 0;
        if (readed > bufsize) readed = bufsize;
        file->seekpos += bufsize;
    }
    else
    {
        pos = (int) (file->seekpos - file->bufstart);
        file->seekpos += bufsize;
        readed = bufsize;
    }

    *buffer = (void *) (file->buffer+pos);
    return readed;
}

int FileBufWrite(FILE_BUF *file, void *buffer, int bufsize)
{
    int readed;
    struct stat statbuf;

    if (fstat(file->handle, &statbuf) != 0) return 0;

    file->bufstart = 1;
    file->bufend = 0;
    //file->last_mtime = statbuf.st_mtime;

    FileSeek(file->handle, file->seekpos, SEEK_SET);
    readed = FileWrite(file->handle, buffer, bufsize);
    file->filesize = lseek(file->handle, 0, SEEK_END);
    if (readed > 0) file->seekpos += readed;

    return readed;
}

long FileBufSeek(FILE_BUF *file, long filepos, int type)
{
    struct stat statbuf;

    if (fstat(file->handle, &statbuf) != 0) return 0xffffffff;

    if (statbuf.st_mtime != file->last_mtime)
    {
        /* File modified, dump buffer */
        file->last_mtime = statbuf.st_mtime;
        file->bufstart = 1;
        file->bufend = 0;
        file->filesize = lseek(file->handle, 0, SEEK_END);
    }

    switch (type)
    {
        case SEEK_SET:
            file->seekpos = filepos;
            break;
        case SEEK_CUR:
            file->seekpos += filepos;
            break;
        case SEEK_END:
            file->seekpos = file->filesize+filepos;
            break;
    }

    return file->seekpos;
}

int erase_path(char *path)
{
    DIR *dirp;
    struct dirent *direntp;
    struct stat statbuf;
    char tmp[256];

    dirp = opendir(path);
    if (dirp == NULL) return 0;

    for (;;)
    {
        direntp = readdir(dirp);
        if (direntp == NULL) break;

        sprintf(tmp, "%s"SSLASH"%s", path, direntp->d_name);
        if (stat(tmp, &statbuf) != 0) continue;
        if (ISDIR(statbuf)) continue;

        remove(tmp);
    }

    closedir(dirp);

    return 1;
}
